from typing import List, Optional, Union

import torch
import numpy as np
from torch.nn.functional import avg_pool2d

from .image_object import Image, Image_List
from astropy.io import fits
from .. import AP_config
from ..errors import SpecificationConflict, InvalidData

__all__ = ["PSF_Image"]


class PSF_Image(Image):
    """Image object which represents a model of PSF (Point Spread Function).

    PSF_Image inherits from the base Image class and represents the model of a point spread function.
    The point spread function characterizes the response of an imaging system to a point source or point object.

    The shape of the PSF data must be odd.

    Attributes:
        data (torch.Tensor): The image data of the PSF.
        psf_upscale (torch.Tensor): Upscaling factor of the PSF. Default is 1.
        identity (str): The identity of the image. Default is None.

    Methods:
        psf_border_int: Calculates and returns the convolution border size of the PSF image in integer format.
        psf_border: Calculates and returns the convolution border size of the PSF image in the units of pixelscale.
        _save_image_list: Saves the image list to the PSF HDU header.
        reduce: Reduces the size of the image using a given scale factor.
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes the PSF_Image class.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
                psf_upscale (int, optional): Upscaling factor of the PSF. Default is 1.
                band (str, optional): The band of the image. Default is None.
        """
        self.psf_upscale = torch.as_tensor(
            kwargs.get("psf_upscale", 1), dtype=torch.int32, device=AP_config.ap_device
        )
        super().__init__(*args, **kwargs)
        self.reference_radec = (0,0)
        self.reference_planexy = (0,0)
        self.reference_imageij = np.flip(np.array(self.data.shape, dtype = float) - 1.) / 2
        self.reference_imagexy = (0,0)

    def set_data(
        self, data: Union[torch.Tensor, np.ndarray], require_shape: bool = True
    ):
        super().set_data(data = data, require_shape = require_shape)
        
        if torch.any(
            (torch.tensor(self.data.shape) % 2) != 1
        ):
            raise SpecificationConflict(f"psf must have odd shape, not {self.data.shape}")
        if torch.any(self.data < 0):
            raise InvalidData("PSF image should have positive values.")
        
    @property
    def psf_border_int(self):
        """Calculates and returns the border size of the PSF image in integer
        format. This is the border used for padding before convolution.

        Returns:
            torch.Tensor: The border size of the PSF image in integer format.

        """
        return torch.ceil(
            (
                1
                + torch.flip(
                    torch.tensor(
                        self.data.shape,
                        dtype=AP_config.ap_dtype,
                        device=AP_config.ap_device,
                    ),
                    (0,),
                )
                / self.psf_upscale
            )
            / 2
        ).int()

    def _save_image_list(self, image_list, psf_header):
        """Saves the image list to the PSF HDU header.

        Args:
            image_list (list): The list of images to be saved.
            psf_header (astropy.io.fits.Header): The header of the PSF HDU.
        """
        psf_header["IMAGE"] = "PSF"
        psf_header["UPSCALE"] = int(self.psf_upscale.detach().cpu().item())
        image_list.append(
            fits.ImageHDU(self.data.detach().cpu().numpy(), header=psf_header)
        )

    def copy(self, **kwargs):
        """Creates a copy of the PSF_Image instance.

        This method generates a copy of the PSF_Image object while maintaining the 'psf_upscale' and 'band' properties.

        Args:
            **kwargs: Arbitrary keyword arguments.

        Returns:
            PSF_Image: A copy of the current PSF_Image instance.
        """
        return super().copy(
            psf_upscale=self.psf_upscale,
        )

    def blank_copy(self, **kwargs):
        """Creates a blank copy of the PSF_Image instance.

        This method generates a blank copy of the PSF_Image object while maintaining the 'psf_upscale' and 'band' properties.

        Args:
            **kwargs: Arbitrary keyword arguments.

        Returns:
            PSF_Image: A blank copy of the current PSF_Image instance with the same properties but no data.
        """
        return super().blank_copy(
            psf_upscale=self.psf_upscale,
        )

    def get_window(self, **kwargs):
        """Returns the window of the PSF_Image instance.

        This method returns the window of the PSF_Image object while maintaining the 'psf_upscale' and 'band' properties.

        Args:
            **kwargs: Arbitrary keyword arguments.

        Returns:
            Window: The window associated with the PSF_Image instance.
        """
        return super().get_window(
            psf_upscale=self.psf_upscale,
        )

    def to(self, dtype=None, device=None):
        """Transfers the PSF_Image instance to the specified device and modifies its data type.

        This method changes the data type of the PSF_Image object and moves it to a specified device.
        If no device is provided, it defaults to 'AP_config.ap_device'.

        Args:
            dtype (torch.dtype, optional): The desired data type to which the PSF_Image instance should be converted.
            device (torch.device, optional): The desired device to which the PSF_Image instance should be moved.
        """
        if device is None:
            device = AP_config.ap_device
        self.psf_upscale = self.psf_upscale.to(device=device)
        super().to(dtype=dtype, device=device)

    def reduce(self, scale, **kwargs):
        """Reduces the size of the image using a given scale factor.

        This method is used to perform a reduction in the size of the PSF image. The new upscaling factor
        is calculated by dividing the existing upscaling factor with the provided scale factor.

        Args:
            scale (float): The scale factor by which the size of the PSF image needs to be reduced.
            **kwargs: Arbitrary keyword arguments. This can be used to pass additional parameters required by the method.

        Returns:
            PSF_Image: A new instance of PSF_Image class with the reduced image size.
        """
        return super().reduce(scale, psf_upscale=self.psf_upscale / scale, **kwargs)

    def expand(self, padding):
        raise NotImplementedError("expand not available for PSF_Image")



import torch

from .star_model_object import Star_Model
from ..image import Model_Image
from ..utils.decorators import ignore_numpy_warnings, default_internal
from ..utils.interpolate import interp2d
from ._shared_methods import select_target
from ..param import Param_Unlock, Param_SoftLimits
from .. import AP_config

__all__ = ["PSF_Star"]


class PSF_Star(Star_Model):
    """Star model which uses an image of the PSF as its representation
    for stars. Using bilinear interpolation it will shift the PSF
    within a pixel to accurately represent the center location of a
    point source. There is no funcitonal form for this object type as
    any image can be supplied. Note that as an argument to the model
    at construction one can provide "psf" as an AstroPhot Model_Image
    object. Since only bilinear interpolation is performed, it is
    recommended to provide the PSF at a higher resolution than the
    image if it is near the nyquist sampling limit. Bilinear
    interpolation is very fast and accurate for smooth models, so this
    way it is possible to do the expensive interpolation before
    optimization and save time. Note that if you do this you must
    provide the PSF as a Model_Image object with the correct PSF
    (essentially just divide the PSF by the upsampling factor you
    used).

    Parameters:
        flux: the total flux of the star model, represented as the log of the total flux.

    """

    model_type = f"psf {Star_Model.model_type}"
    parameter_specs = {
        "flux": {"units": "log10(flux/arcsec^2)"},
    }
    _parameter_order = Star_Model._parameter_order + ("flux",)
    useable = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # fixme, model already has PSF interface, those can just be merged
        if "psf" in kwargs:
            self.psf_model = kwargs["psf"]
        else:
            self.psf_model = Model_Image(
                data=torch.clone(self.psf.data),
                pixelscale=self.psf.pixelscale,
            )
        self.psf_model.header.shift(
            self.psf_model.origin - self.psf_model.center
        )

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)
        target_area = target[self.window]
        with Param_Unlock(parameters["flux"]), Param_SoftLimits(parameters["flux"]):
            if parameters["flux"].value is None:
                parameters["flux"].value = torch.log10(torch.abs(torch.sum(target_area.data)) / target.pixel_area)
            if parameters["flux"].uncertainty is None:
                parameters["flux"].uncertainty = torch.abs(parameters["flux"].value) * self.default_uncertainty

    @default_internal
    def evaluate_model(self, X=None, Y=None, image=None, parameters=None, **kwargs):
        if X is None:
            Coords = image.get_coordinate_meshgrid()
            X, Y = Coords - parameters["center"].value[..., None, None]

        # Convert coordinates into pixel locations in the psf image
        pX, pY = self.psf_model.plane_to_pixel(X, Y)
        pX = pX.reshape(X.shape)
        pY = pY.reshape(Y.shape)

        # Select only the pixels where the PSF image is defined
        select = torch.logical_and(
            torch.logical_and(pX > -0.5, pX < self.psf_model.data.shape[1]),
            torch.logical_and(pY > -0.5, pY < self.psf_model.data.shape[0]),
        )

        # Zero everywhere outside the psf
        result = torch.zeros_like(X)

        # Use bilinear interpolation of the PSF at the requested coordinates
        result[select] = interp2d(self.psf_model.data, pX[select], pY[select])

        return result * (image.pixel_area * 10 ** parameters["flux"].value)



import torch

from .model_object import Component_Model
from ..utils.decorators import default_internal


__all__ = ["Star_Model"]


class Star_Model(Component_Model):
    """Prototype star model, to be subclassed by other star models which
    define specific behavior. Mostly this just requires that no
    standard PSF convolution is applied to this model as that is to be
    handled internally by the star model. By default, the PSF object
    has no integration mode since it will always be evaluated at the
    resolution of the PSF provided, the integration has in effect
    already been done when constructing the PSF.

    """

    model_type = f"star {Component_Model.model_type}"
    useable = False

    @property
    def psf_mode(self):
        return "none"

    @psf_mode.setter
    def psf_mode(self, val):
        pass



from typing import Optional, Union, Dict, Tuple, Any
from copy import deepcopy

import numpy as np
import torch

from ..param import Parameter_Node
from ..utils.decorators import ignore_numpy_warnings, default_internal
from ..utils.interpolate import (
    _shift_Lanczos_kernel_torch,
    simpsons_kernel,
    curvature_kernel,
    interp2d,
)
from ..image import Model_Image, Target_Image, Window
from ..utils.operations import (
    fft_convolve_torch,
    fft_convolve_multi_torch,
    grid_integrate,
)
from .. import AP_config


@default_internal
def angular_metric(self, X, Y, image=None, parameters=None):
    return torch.atan2(Y, X)


@default_internal
def radius_metric(self, X, Y, image=None, parameters=None):
    return torch.sqrt(
        (X) ** 2 + (Y) ** 2 + self.softening**2
    )


@classmethod
def build_parameter_specs(cls, user_specs=None):
    parameter_specs = {}
    for base in cls.__bases__:
        try:
            parameter_specs.update(base.build_parameter_specs())
        except AttributeError:
            pass
    parameter_specs.update(cls.parameter_specs)
    parameter_specs = deepcopy(parameter_specs)
    if isinstance(user_specs, dict):
        for p in user_specs:
            # If the user supplied a parameter object subclass, simply use that as is
            if isinstance(user_specs[p], Parameter_Node):
                parameter_specs[p] = user_specs[p]
            elif isinstance(
                user_specs[p], dict
            ):  # if the user supplied parameter specifications, update the defaults
                parameter_specs[p].update(user_specs[p])
            else:
                parameter_specs[p]["value"] = user_specs[p]

    return parameter_specs


def build_parameters(self):
    for p in self.__class__._parameter_order:
        # skip if the parameter already exists
        if p in self.parameters:
            continue
        # If a parameter object is provided, simply use as-is
        if isinstance(self.parameter_specs[p], Parameter_Node):
            self.parameters.link(self.parameter_specs[p].to())
        elif isinstance(self.parameter_specs[p], dict):
            self.parameters.link(Parameter_Node(p, **self.parameter_specs[p]))
        else:
            raise ValueError(f"unrecognized parameter specification for {p}")


def _sample_init(self, image, parameters, center):
    if self.sampling_mode == "midpoint" and max(image.data.shape) >= 100:
        Coords = image.get_coordinate_meshgrid()
        X, Y = Coords - center[..., None, None]
        mid = self.evaluate_model(X=X, Y=Y, image=image, parameters=parameters)
        kernel = curvature_kernel(AP_config.ap_dtype, AP_config.ap_device)
        # convolve curvature kernel to numericall compute second derivative
        curvature = torch.nn.functional.pad(
            torch.nn.functional.conv2d(
                mid.view(1, 1, *mid.shape),
                kernel.view(1, 1, *kernel.shape),
                padding="valid",
            ),
            (1, 1, 1, 1),
            mode="replicate",
        ).squeeze()
        return mid + curvature, mid
    elif self.sampling_mode == "trapezoid" and max(image.data.shape) >= 100:
        Coords = image.get_coordinate_corner_meshgrid()
        X, Y = Coords - center[..., None, None]
        dens = self.evaluate_model(X=X, Y=Y, image=image, parameters=parameters)
        kernel = (
            torch.ones(
                (1, 1, 2, 2), dtype=AP_config.ap_dtype, device=AP_config.ap_device
            )
            / 4.0
        )
        trapz = torch.nn.functional.conv2d(
            dens.view(1, 1, *dens.shape), kernel, padding="valid"
        )
        trapz = trapz.squeeze()
        kernel = curvature_kernel(AP_config.ap_dtype, AP_config.ap_device)
        curvature = torch.nn.functional.pad(
            torch.nn.functional.conv2d(
                trapz.view(1, 1, *trapz.shape),
                kernel.view(1, 1, *kernel.shape),
                padding="valid",
            ),
            (1, 1, 1, 1),
            mode="replicate",
        ).squeeze()
        return trapz + curvature, trapz

    Coords = image.get_coordinate_simps_meshgrid()
    X, Y = Coords - center[..., None, None]
    dens = self.evaluate_model(X=X, Y=Y, image=image, parameters=parameters)
    kernel = simpsons_kernel(dtype=AP_config.ap_dtype, device=AP_config.ap_device)
    # midpoint is just every other sample in the simpsons grid
    mid = dens[1::2, 1::2]
    simps = torch.nn.functional.conv2d(
        dens.view(1, 1, *dens.shape), kernel, stride=2, padding="valid"
    )
    return mid.squeeze(), simps.squeeze()


def _integrate_reference(self, image_data, image_header, parameters):
    return torch.sum(image_data) / image_data.numel()


def _sample_integrate(self, deep, reference, image, parameters, center):
    if self.integrate_mode == "none":
        pass
    elif self.integrate_mode == "threshold":
        Coords = image.get_coordinate_meshgrid()
        X, Y = Coords - center[..., None, None]
        ref = self._integrate_reference(
            deep, image.header, parameters
        )  # fixme, error can be over 100% on initial sampling reference is invalid
        error = torch.abs((deep - reference))
        select = error > (self.sampling_tolerance * ref)
        intdeep = grid_integrate(
            X=X[select],
            Y=Y[select],
            image_header=image.header,
            eval_brightness=self.evaluate_model,
            eval_parameters=parameters,
            dtype=AP_config.ap_dtype,
            device=AP_config.ap_device,
            quad_level=self.integrate_quad_level,
            gridding=self.integrate_gridding,
            max_depth=self.integrate_max_depth,
            reference=self.sampling_tolerance * ref,
        )
        deep[select] = intdeep
    else:
        raise ValueError(
            f"{self.name} has unknown integration mode: {self.integrate_mode}"
        )
    return deep


def _sample_convolve(self, image, shift, psf, shift_method="bilinear"):
    """
    image: Image object with image.data pixel matrix
    shift: the amount of shifting to do in pixel units
    psf: a PSF_Image object
    """
    if shift is not None:
        if shift_method == "bilinear":
            psf_data = torch.nn.functional.pad(psf.data, (1, 1, 1, 1))
            X, Y = torch.meshgrid(
                torch.arange(
                    psf_data.shape[1],
                    dtype=AP_config.ap_dtype,
                    device=AP_config.ap_device,
                )
                - shift[0],
                torch.arange(
                    psf_data.shape[0],
                    dtype=AP_config.ap_dtype,
                    device=AP_config.ap_device,
                )
                - shift[1],
                indexing="xy",
            )
            shift_psf = interp2d(psf_data, X.clone(), Y.clone())
        elif "lanczos" in shift_method:
            lanczos_order = int(shift_method[shift_method.find(":") + 1 :])
            psf_data = torch.nn.functional.pad(
                psf.data, (lanczos_order, lanczos_order, lanczos_order, lanczos_order)
            )
            LL = _shift_Lanczos_kernel_torch(
                -shift[0],
                -shift[1],
                lanczos_order,
                AP_config.ap_dtype,
                AP_config.ap_device,
            )
            shift_psf = torch.nn.functional.conv2d(
                psf_data.view(1, 1, *psf_data.shape),
                LL.view(1, 1, *LL.shape),
                padding="same",
            ).squeeze()
        else:
            raise ValueError(f"unrecognized subpixel shift method: {shift_method}")
    else:
        shift_psf = psf.data
    shift_psf = shift_psf / torch.sum(shift_psf)
    if self.psf_convolve_mode == "fft":
        image.data = fft_convolve_torch(image.data, shift_psf, img_prepadded=True)
    elif self.psf_convolve_mode == "direct":
        image.data = torch.nn.functional.conv2d(
            image.data.view(1, 1, *image.data.shape),
            torch.flip(
                shift_psf.view(1, 1, *shift_psf.shape),
                dims=(2, 3),
            ),
            padding="same",
        ).squeeze()
    else:
        raise ValueError(f"unrecognized psf_convolve_mode: {self.psf_convolve_mode}")
